from BlocksWorld import *
goalStatePosSet = list()
    
def h1(n) :
    """An efficient heuristic on the search of an optimal solution. It uses a list(goalStatePosSet) containing
    sets of the goal state. The index of each set inside the list is the position of the elements on their
    respective lists. For example in a goal state that contains two lists A,B,C and D,E the goalStatePosSet
    will contain three sets, the first at index 0 with A and D, the second at index 1 with B and E, and
    the third at index 2 with C. This list is created before the search starts. Each element is a tuple of the
    previous cube in the list and the name of the cube. If it's the first one the previous var is -1.

    The logic behind this function is to check all the cube lists of the state bottom-up and stop
    where it doesn't find a match compared to the goal state. It adds then the remaining cubes of the
    list to the estimation of the action's cost. It also gives a little advantage when the choice is difficult
    to the actions that create new lists, expanding the cubes for flexibility.

    Returns the estimation of the cost of the action."""
    estim = 0
    
    for i in range(len(n.state)) :
        prev = -1
        for j in range(len(n.state[i])) :
            if j > 0 :
                prev = n.state[i][j-1]
            if j >= len(goalStatePosSet) or (prev, n.state[i][j]) not in goalStatePosSet[j]:
                estim += len(n.state[i]) - j
                break

    if n.parent != None :
        if len(n.state) > len(n.parent.state) :
            estim -= 0.5
    
    return estim

#_______________________________________________________________


input = open('input.txt', 'r') #Input provided by a file called input.txt as described in the README.txt

for line in input :
    start_time = time.time()
    
    line = line.rstrip('\n')

    initAndGoal= line.split(">")
    initialState = initAndGoal[0]
    goalState = initAndGoal[1]

    initialState= initialState.split("|")
    for i in range(len(initialState)) :
        initialState[i] = initialState[i].split(",")
        initialState[i] = tuple(initialState[i])

    goalState= goalState.split("|")
    for i in range(len(goalState)) :
        goalState[i] = goalState[i].split(",")
        goalState[i] = tuple(goalState[i])
        
    initialState = tuple(initialState)
    goalState = tuple(goalState)
    
    print "Initial state: {0}\nGoal state: {1}".format(initialState, goalState)

    #Construction of the goalStatePosSet as explained at the description of the h1
    for i in range(len(goalState)) :
        prev = -1
        for j in range(len(goalState[i])) :
            if j > 0 :
                prev = goalState[i][j-1]
            if len(goalStatePosSet) <= j :
                goalStatePosSet.insert(j, set())
                goalStatePosSet[j].add((prev, goalState[i][j]))
            else :
                goalStatePosSet[j].add((prev, goalState[i][j]))
    
    p = BlocksWorld(initialState, goalState)
    s = astar_search(p, h1)
    sol = s.solution() # The sequence of actions to go from the root to this node
    path = s.path() # The nodes that form the path from the root to this node

    print "Solution: \n"
    
    for i in range(len(path)) :
        state = path[i].state
        cost = path[i].path_cost
        action = " "
        if i > 0 : # The initial state has not an action that results to it
            action = sol[i-1]
        print "Action {0}: {1}\nState {0}: {2}\nPath Cost: {3}\n".format(i, action, state, cost)

    print "{1}\nTime elapsed: {0} sec\n{1}\n".format(time.time() - start_time, '-'*60)
    
    goalStatePosSet = list()
